// Wrapper functions for linear algebra routines
// Kenji Sugase		12/13/2013
// Dan M.      		11/2023

#include <iostream>
#include <cmath>
#include "LinearAlgebra.h"
#include "Alloc.h"
#include <complex.h>
#include <math.h>
#include <cmath>    // round etc

using namespace std;

int idxr(int c, int r, int order) { return 2*(c + order*r);     }
int idxi(int c, int r, int order) { return 2*(c + order*r) + 1; }
int idvr(int c) { return 2*c;     }
int idvi(int c) { return 2*c + 1; }

// copy a square matrix from one place to another
void zmat_cpy( int order, const cdouble *A, cdouble *B) {
	zlacpy_("A", &order, &order, A, &order, B, &order);
}

// Compute an inverse matrix using LU decomposition: A-1
int zmat_inv(int order, const cdouble *mat, cdouble *imat) {
	int lwork = 2*order;
	int ipiv[order];
	cdouble work[4*order];
	// Duplicate the matrix A
	zlacpy_("A", &order, &order, mat, &order, imat, &order);
	// LU decomposition
	int info;
	zgetrf_(&order, &order, imat, &order, ipiv, &info);
	if (info) {
		cerr << "WARNING: Relaxation matrix is singular. Check the input file.\n";
	} else {
		// Compute the inverse of A
		zgetri_(&order, imat, &order, ipiv, work, &lwork, &info);
		if (info) cerr << "WARNING: Relaxation matrix is singular. Check the input file.\n";
	}
	return info;
}

// C = A x B
void zmat_mlt(int order, const cdouble *A, const cdouble *B, cdouble *C) {
        double alpha[2] = {1.0, 0.0};
        double beta[2]  = {0.0, 0.0};
        zgemm_("N", "N", &order, &order, &order, alpha, A, &order, B, &order, beta, C, &order);
}

// D = A * B * C
void zmat_mlt(int order, const cdouble *A, const cdouble *B, const cdouble *C, cdouble *D) {
        cdouble tmp[2*order*order];
	zmat_mlt(order, A, B, tmp);
        zmat_mlt(order, tmp, C, D);
}

// calc eigenvalues and right eigenvector
int zmat_eig(int order, cdouble *mat, cdouble *evec, cdouble *eval) {
        int lwork = 2*order;
        double rwork[2*order];
        cdouble work[4*order];
        cdouble *vl = NULL;
        int ldvl = 1;
        int info;
        zgeev_("N", "V", &order, mat, &order, eval,  
		vl, &ldvl, evec, &order, 
		work, &lwork, rwork, &info);

        if (info) {
                cerr << "\nERROR: Eigenvectors of the relaxation matrix cannot be computed.\n";
                exit(1);
        }
        return info;
}

int dmat_eig(int order, double *mat, double *eigReal, double *eigImag) {
  int n = order;
  char Nchar='N';
//  double *eigReal=new double[n];
//  double *eigImag=new double[n];
  double *vl,*vr;
  int one=1;
  int lwork=6*n;
  double *work=new double[lwork];
  int info;

   	dgeev_(&Nchar,&Nchar,&order,mat,&order,eigReal,eigImag,
            vl,&one,vr,&one,
            work,&lwork,&info);
        if (info) {
                cerr << "\nERROR: Eigenvectors of the relaxation matrix cannot be computed.\n";
                exit(1);
        }
        return info;
}

// calc eigenvalues and right eigenvector
int zmat_evv(int order, cdouble *mat, cdouble *vr, cdouble *vl, cdouble *eval) {
        int lwork = 2*order;
        double rwork[2*order];
        //double lwork[2*order];
        cdouble work[4*order];
        int info;
        zgeev_("V", "V", &order, mat, &order, eval,
                vl, &order, vr, &order,
                work, &lwork, rwork, &info);

        if (info) {
                cerr << "\nERROR: Eigenvectors of the relaxation matrix cannot be computed.\n";
                exit(1);
        }
        return info;
}

// Compute a conjugate complex transpose matrix: A*
void zmat_cnj(int order, const cdouble *A, cdouble *cnjA) {
        for (int n = 0; n < 2*order*order ; n += 1) {
		int i = n/order ; 
		int j = n%order ; 
		int k = order*j+i ;
                cnjA[n]   =  conj(A[k]);
   }
}

// Add matrix A + B = C
void zmat_add(int order, const cdouble *A, cdouble *B, cdouble *C) {
        for (int n = 0; n < 2*order*order ; n += 1) {
                C[n]   =  A[n] + B[n] ;
   }
}

// Add matrix A + B + C = D
void zmat_add(int order, const cdouble *A, cdouble *B, cdouble *C, cdouble *D) {

        cdouble tmp[2*order*order];
	zmat_add(order, A, B, tmp);
        zmat_add(order, tmp, C, D);
}

// round matrix A => B
void zmat_rnd(int order, cdouble *A, cdouble *B ) {
        for (int n = 0; n < 2*order*order ; n += 1) {
                B[n]   =  {round(real(A[n])), round(imag(A[n]))} ;
   }
}


// B = exp( t * eigVal ) // diagonal matrix only!
void zmat_exp(int order, double t, const cdouble *A, cdouble *B) {
        for (int n = 0; n < 2*order*order ; n += 1) {
		
		if ( n%(order+1) == 0 ) {
                	B[n]   =  exp( t * A[n]) ;	
		}
		else { 	B[n]   = {0. , 0.} ; }
   }
}

// Computes Y = A * X for matrix A and vector X
// need "T" transpose call to get it right row*col convention 
void zmat_vmlt(int order, const cdouble *A, const cdouble *X, cdouble *Y) {
        cdouble alpha[2] = {1.0,0.0};
        cdouble beta[2] = {0.0,0.0};
        int lda = order;
        int incx = 1;
        int incy = 1;
        zgemv_("T", &order, &order, alpha, A, &lda, X, &incx, beta, Y, &incy);
}

// C = A x B  (row vector * Matrix)
void zmat_rvm(int order, const cdouble *A, const cdouble *B, cdouble *C) {
        double alpha[2] = {1.0, 0.0};
        double beta[2]  = {0.0, 0.0};
	int one = 1 ;
        zgemm_("N", "T", &one, &order, &order, alpha, A, &one, B, &order, beta, C, &one );
}

// Computes Y = A * X  (where A = row vector and X = column vector)
// need "T" transpose call to get it right row*col convention 
void zmat_rvv(int order, const cdouble *A, const cdouble *X, cdouble *Y) {
        cdouble alpha[2] = {1.0,0.0};
        cdouble beta[2] = {0.0,0.0};
        int lda = order;
        int incx = 1;
        int incy = 1;
        int one  = 1;
        zgemv_("T", &order, &one, alpha, A, &lda, X, &incx, beta, Y, &incy);
}








